from collections import Counter
from re import finditer, compile
from typing import List
import inflect
import os
import boto3

# from functools import lru_cache

inflect_engine = inflect.engine()  # used for singular/plural word comparing


_camel_case_regex_compiled = compile(
    r".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)"
)


def _create_boto_session():
    profile_name = os.environ.get("AWS_PROFILE", False)
    region_name = os.environ.get("AWS_REGION", False)
    _kwargs_dict = {}
    if profile_name:
        _kwargs_dict["profile_name"] = profile_name
    if region_name:
        _kwargs_dict["region_name"] = region_name
    session = boto3.session.Session(**_kwargs_dict)
    return session


def are_two_lists_same(list_one: List, list_two: List) -> bool:
    """Compares the contents of two lists"""
    return Counter(list_one) == Counter(list_two)


def is_word_in_a_list_of_words(word: str, list_of_words: List[str]) -> bool:
    """Checks if a word is a is in a list_of_words, case insensitive"""
    lower_word = word.lower()
    for a_word in list_of_words:
        if lower_word == a_word.lower():
            return True
    return False

    return word.lower() in []


def ifind_similar_names_in_list(
    search_for: str, search_in_list: List[str]
) -> List[str]:
    """Case insensitive find in list"""
    result = []
    if not search_for or not search_in_list:
        return result
    lower_search_for = search_for.lower()

    for l_name in search_in_list:
        if l_name.lower().startswith(lower_search_for):
            result.append(l_name)
    return result


def str_relations(relations: List[dict]) -> str:
    """Stringify list of relations"""
    return ", ".join(
        [f"[{r.service_name}.{r.operation_name} > {r.target_path}]" for r in relations]
    )


def get_all_available_services(session) -> List[str]:
    """Gets available services from boto3 session"""
    return session.get_available_services()


def compare_nouns(word1: str, word2: str) -> bool:
    """Singular/plural insensitive word comparison"""
    return inflect_engine.compare_nouns(word1, word2)


def compare_two_token_lists(
    token_list_one: List[str], token_list_two: List[str]
) -> bool:
    """Compares to word lists one by one in lowercase"""
    if len(token_list_one) != len(token_list_two):
        return False
    token_by_token_match = [
        w1.lower() == w2.lower() for w1, w2 in zip(token_list_one, token_list_two)
    ]
    has_token_by_token_match = all(token_by_token_match)
    return has_token_by_token_match


def icompare_two_token_lists(
    token_list_one: List[str], token_list_two: List[str]
) -> bool:
    """Singular/plural insensitive token list comparison"""
    if len(token_list_one) != len(token_list_two):
        return False
    token_by_token_match = [
        bool(inflect_engine.compare_nouns(w1.lower(), w2.lower()))
        for w1, w2 in zip(token_list_one, token_list_two)
    ]
    has_token_by_token_match = all(token_by_token_match)
    return has_token_by_token_match


def camel_case_split(identifier: str) -> List[str]:
    """Splits CamelCase string to it's tokens"""
    matches = finditer(_camel_case_regex_compiled, identifier)
    return [m.group(0) for m in matches]


def compare_two_camel_case_words(word1: str, word2: str) -> bool:
    token_list_1 = [_.lower() for _ in camel_case_split(word1)]
    token_list_2 = [_.lower() for _ in camel_case_split(word2)]
    return compare_two_token_lists(token_list_1, token_list_2)


def icompare_two_camel_case_words(word1: str, word2: str) -> bool:
    token_list_1 = [_.lower() for _ in camel_case_split(word1)]
    token_list_2 = [_.lower() for _ in camel_case_split(word2)]
    return icompare_two_token_lists(token_list_1, token_list_2)



from botocore_utils import (
    _flatten_shape_to_its_members_and_target_paths,
    get_shape_name,
)
from utils import (
    camel_case_split,
    compare_nouns,
    icompare_two_token_lists,
    compare_two_camel_case_words,
    str_relations,
    is_word_in_a_list_of_words
)
from botocore_utils import (
    get_input_shape,
    get_max_results_value_from_shape,
    find_key_in_dict_keys,
    generate_rich_tree_from_shape,
    icompare_two_camel_case_words,
    ifind_key_in_dict_keys,
    READ_ONLY_VERBS,
    IDENTIFIER_NAMES,
    cleanhtml,
)
from relations import RelationMap, Relation
from reader import ServiceReader
from registries import ResourceNodeRegistry
from config import get_logger, get_rich_console
from errors import Error

from typing import List, Dict, Tuple, Union
from botocore.utils import ArgumentGenerator
from botocore.hooks import EventAliaser
from botocore.model import OperationModel, ServiceModel
from rich.text import Text
from rich.panel import Panel
from rich.console import Group
from rich.padding import Padding
import jmespath

logger = get_logger(__name__)

_resource_node_registry = ResourceNodeRegistry()
argument_generator = ArgumentGenerator()
console = get_rich_console()
PAGINATION_TOKEN_KEYS = ("nexttoken", "continuationtoken", "marker")

class ResourceNode:
    def __init__(
        self, service_node: "ServiceNode", name: str, operation_names: List[str]
    ) -> None:
        self.service_node = service_node
        self.name = name
        self.operation_names = operation_names
        self._operation_models = {}

    def __init_subclass__(cls, service_name=None, name=None, **kwargs) -> None:
        """Initializes the custom subclasses of ResourceNode to ResourceNodeRegistry.

        Args:
            service_name (_type_, optional): Name of the AWS service. Defaults to None.
            name (_type_, optional): Name of the AWS Resource Node. Defaults to None.
        """
        super().__init_subclass__(**kwargs)
        if service_name and name:
            _resource_node_registry.register_class(cls, service_name, name)
        else:
            logger.debug(
                f'{cls.__name__} invalid! You must define "service_name" and "name"'
            )

    def get_operation_types_and_names(self) -> Dict[str, str]:
        operation_names = self.get_operation_names()
        types_to_operation_names = {}
        for op_name in operation_names:
            verb, *resource_node_name_tokens = camel_case_split(op_name)
            resource_node_name = "".join(resource_node_name_tokens)
            if verb.lower() == "describe":
                types_to_operation_names["describe"] = op_name
            if verb.lower() == "list":
                types_to_operation_names["list"] = op_name

            if verb.lower() == "get":
                if self.name != resource_node_name:
                    types_to_operation_names["gets"] = op_name
                else:
                    types_to_operation_names["get"] = op_name
        return types_to_operation_names

    def get_pagination_output_to_input_keys(self) -> Dict[str, str]:
        """

        For example, ContinuationToken:

        Returns:
            Dict[str, str]: Single k/v pair dictionary.
        """
        return {}

    def get_operation_names(self) -> List[str]:
        """Returns the available operation names in the ResourceNode.

        Returns:
            List[str]: Operation names tied to the ResourceNode
        """
        return self.operation_names

    # NOTE: +overrideable
    def define_extra_relations(self) -> Union[List[Dict], List[Relation]]:
        """Extra relations defined in the custom subclasses of ResourceNode

        Returns:
            Union[List[Dict], List[Relation]]: List of relations as dicts or Relation objects
        """
        return []

    # NOTE: +overrideable
    def get_operations_relations(
        self, operation_name: str
    ) -> Tuple[Union[List[Dict], bool], Union[Error, None]]:
        """_summary_

        Args:
            operation_name (str): Name of the operation.

        Returns:
            Tuple[Union[List[Dict], bool], Union[Error, None]]: Returns value and error.
                                                                True: No required parameters
                                                                False: Failure
                                                                List[Dict]: List of relations
        """

        resource_node = self
        relation_map = self.service_node.get_relation_map()
        required_parameter_names = (
            resource_node.get_required_parameter_names_from_operation_name(
                operation_name
            )
        )
        # required_parameter_names_to_relations_map = {
        #     r_param_name: None for r_param_name in required_parameter_names
        # }

        if not required_parameter_names:
            # no required parameters
            return True, None

        operation_markup = (
            f"[bold][green]{self.service_node.name}[/].[blue]{operation_name}[/][/]"
        )
        req_param_markup = f"[bold magenta]{', '.join(required_parameter_names)}[/]"

        if len(required_parameter_names) == 1:

            # only one parameter exists
            single_relation_list = None
            single_parameter_name = required_parameter_names[0]
            generated_relations_for_parameter = (
                relation_map.get_parameters_generated_relations(
                    single_parameter_name, operation_name
                )
            )
            if not generated_relations_for_parameter:
                logger.debug(
                    f"Failed to generate relations. {operation_markup} has a required parameters: {req_param_markup}"
                )
                return False, Error(
                    "failed to generate relations",
                    {
                        "required_parameter_names": required_parameter_names,
                        "service": self.service_node.name,
                        "resource_node": self.name,
                        "operation_name": operation_name,
                    },
                )

            single_relation_list = (
                resource_node.find_best_relation_for_single_parameter(
                    single_parameter_name, generated_relations_for_parameter
                )
            )
            if single_relation_list:
                logger.debug(
                    f"[green]Success finding relations.[/] {operation_markup} has a required parameter: {req_param_markup}. Relation found: [yellow]{str_relations(single_relation_list)}[/]"
                )
                return single_relation_list, None
            else:
                logger.debug(
                    f"Failed to choose the best relation for operation. {operation_markup} has required parameters: {req_param_markup}"
                )
                return False, Error(
                    "failed to choose the best relation",
                    {
                        "service": self.service_node.name,
                        "resource_node": self.name,
                        "operation_name": operation_name,
                        "generated_relations_for_parameter": generated_relations_for_parameter,
                    },
                )
        else:
            # multiple required parameters does not supported
            return False, Error(
                "multiple parameters - must be extended with a subclass",
                {
                    "service": self.service_node.name,
                    "resource_node": self.name,
                    "operation_name": operation_name,
                },
            )

    # NOTE: +overrideable
    def generate_jmespath_selector_from_relations(
        self, operation_name: str, relation_list: List[Dict]
    ) -> str:
        """
        Tries to generate the jmespath selector string from given relations. Could be overridden with custom subclasses.

        Args:
            operation_name (str): Name of the operation
            relation_list (List[Dict]): List of relations for the operation. Used to generate the jmespath selector.

        Returns:
            str: Jmespath selector string
        """
        if not len(relation_list) == 1:
            return False
        relation = relation_list[0]

        target_path = relation.target_path
        search_shape = relation.required_shape_name

        _for_all_the_responses = ""
        if not target_path.startswith("[*]."):
            _for_all_the_responses = "[*]."
        _flatten_two_times = "[][]"
        if not relation_list:
            return False

        if "." not in target_path:
            return False

        before_last_attr, last_attribute = target_path.rsplit(".", maxsplit=1)
        # name switch
        jmespath_curly_name_transform = f"{search_shape}: {last_attribute}"

        if not before_last_attr:
            before_last_attr = "[*]"

        # patching in the first part
        jmespath_selector = (
            before_last_attr + ".{" + jmespath_curly_name_transform + "}"
        )
        flattened_jmespath_nested_selector = (
            f"{_for_all_the_responses}{jmespath_selector}{_flatten_two_times}"
        )
        return flattened_jmespath_nested_selector

    # NOTE: +overrideable
    def _generate_raw_api_parameters_from_operation_data(
        self,
        operation_name: str,
        relations_of_operation: List[dict],
        related_operations_data: Union[Dict, List],
    ) -> List:
        """Generates the jmespath selector and search the data with it. Output is the list of api_parameters dictionaries.
        Only considers the required parameters(raw parameters) of the operation.
        > Note: Pagination parameters (e.g. `NextToken`, `MaxResults`) are not included.

        Args:
            operation_name (str): Name of the operation
            relations_of_operation (List[dict]): Relations of the operation
            related_operations_data (Union[Dict, List]): All related operations data

        Returns:
            List: List of required _raw_ API parameters, not including pagination parameters
        """
        if len(relations_of_operation) != 1:
            logger.debug(
                f"Not supported: jmespath selector genereation for multiple relations: {operation_name}"
            )
            return False, Error(
                "multiple parameters - must be extended with a subclass",
                {
                    "service": self.service_node.name,
                    "resource_node": self.name,
                    "operation_name": operation_name,
                },
            )

        direct_relation = relations_of_operation[0]
        direct_relation_op_markup = f"[bold][green]{direct_relation.resource_node_name}[/].[blue]{direct_relation.operation_name}[/][/]"

        resource_node = self
        generated_jmespath_nested_selector = (
            resource_node.generate_jmespath_selector_from_relations(
                operation_name, relations_of_operation
            )
        )
        if not generated_jmespath_nested_selector:
            logger.debug(
                f"Failed to generate JMESPATH selector for [bold][blue][{resource_node.name}[/].[green]{operation_name}[/]]"
            )
            return False, Error(
                "failed to generate jmespath selector",
                {
                    "service": self.service_node.name,
                    "resource_node": self.name,
                    "operation_name": operation_name,
                },
            )
        logger.debug(
            f"Successfuly generated JMESPATH Selector: '[bold]{generated_jmespath_nested_selector}[/]' to extract from {direct_relation_op_markup}"
        )

        if not related_operations_data:
            logger.debug(
                "Failed to find related operations data. Can't generate api parameters with the generated JMESPATH Selector"
            )

        # the first relation we will get
        direct_related_operation = direct_relation.operation_name
        directly_related_operation_data = related_operations_data.get(
            direct_related_operation
        )
        found_api_paramaters = jmespath.search(
            generated_jmespath_nested_selector, directly_related_operation_data
        )
        raw_api_parameters_list = found_api_paramaters
        # for r_api_param in found_api_paramaters:
        #     for r_api_p_value in r_api_param.values():
        #         if bool(r_api_p_value):
        #             raw_api_parameters_list.append(r_api_param)
        if raw_api_parameters_list == []:
            # successfull jmespath search that yielded no results. operation data might be empty
            return raw_api_parameters_list, Error(
                "related resources not found",
                {
                    "service": self.service_node.name,
                    "resource_node": self.name,
                    "operation_name": operation_name,
                },
            )
        elif not raw_api_parameters_list:
            logger.debug(
                f"CANT GENERATE API PARAMETERS LIST WITH [bold][red]{generated_jmespath_nested_selector}[/] {related_operations_data=}"
            )
            return False, Error(
                "failed to generate api parameters",
                {
                    "service": self.service_node.name,
                    "resource_node": self.name,
                    "operation_name": operation_name,
                },
            )

        return raw_api_parameters_list, None

    # NOTE: +overrideable
    def complement_api_parameters_list(
        self,
        operation_name: str,
        related_operations_data: Union[List, Dict],
        relations_of_operation: List[Dict],
        raw_api_parameters_list: List,
    ) -> List[Dict]:
        """Uses the `raw_api_parameters_list` and appends pagination parameters(MaxResults,...) to them.
            Also provided for easy subclass overriding.

        Args:
            operation_name (str): Name of the operation.
            related_operations_data (Union[List, Dict]): All related operations data
            relations_of_operation (List[Dict]): Relations of the operation
            raw_api_parameters_list (List): Generated raw API parameters

        Returns:
            List[Dict]: Valid API parameters to call the boto operation with
        """

        operation_model = self.get_operation_model(operation_name)
        required_parameter_names = (
            self.get_required_parameter_names_from_operation_model(operation_model)
        )

        input_shape = get_input_shape(operation_model)
        generated = {}

        # input shape may not exists, meaning no required parameters
        if input_shape:
            generated = argument_generator.generate_skeleton(input_shape)

        api_params = [{}]
        if raw_api_parameters_list:
            api_params = raw_api_parameters_list

        # handle MaxResults
        max_results_value = get_max_results_value_from_shape(input_shape)
        max_results_key = find_key_in_dict_keys("maxresults", generated)
        if max_results_key and max_results_value:
            for api_param in api_params:
                api_param.update({max_results_key: max_results_value})

        # sometimes MaxResults can be seen as non required parameter, but in fact is
        if not required_parameter_names:
            return api_params

        if raw_api_parameters_list == False:
            # this func expect raw_api_parameters_list present, only adds to prepared parameters
            logger.debug(
                f"FAILED TO CREATE VALID API PARAMETERS. Required Parameters are: [bold]{required_parameter_names}[/]"
            )
            return False

        return api_params

    # NOTE: +overrideable
    def generate_api_parameters_from_operation_data(
        self,
        operation_name: str,
        relations_of_operation: List[Dict],
        related_operations_data: Union[List, Dict],
    ) -> Tuple[Union[List, bool], Union[Error, None]]:
        """Generates API parameters for the given operation including pagination parameters.

        Args:
            operation_name (str): Name of the operation
            relations_of_operation (List[Dict]): Relations of the operation
            related_operations_data (Union[List, Dict]): All related operations data

        Returns:
            List: Generated API Parameters to call the Operations with
        """
        resource_node = self
        required_parameters = (
            resource_node.get_required_parameter_names_from_operation_name(
                operation_name
            )
        )
        no_relations = relations_of_operation == []
        no_required_parameters = required_parameters == []

        if no_required_parameters:
            # Even though we know there are no required parameters,
            # some parameters like MaxResults must be filled, if available.
            api_parameters = resource_node.complement_api_parameters_list(
                operation_name, related_operations_data, [], []
            )
            return api_parameters, None

        if no_relations:
            # todo
            pass
        # try to automatically generate the required parameters with relations
        (
            raw_api_parameters_list,
            raw_param_error,
        ) = self._generate_raw_api_parameters_from_operation_data(
            operation_name, relations_of_operation, related_operations_data
        )
        if raw_param_error is not None:
            # failed to generate raw api parameters list
            return False, raw_param_error

        # fill out the pagination parameters like MaxResults, ...
        api_parameters_list = resource_node.complement_api_parameters_list(
            operation_name,
            related_operations_data,
            relations_of_operation,
            raw_api_parameters_list,
        )
        if not api_parameters_list:
            return api_parameters_list, Error(
                "failed to generate api parameters",
                {
                    "service": self.service_node.name,
                    "resource_node": self.name,
                    "operation_name": operation_name,
                },
            )

        return api_parameters_list, None

    def print_operation(self, operation_name: str) -> None:
        operation_panel = self._rich_operation_details_panel(operation_name)
        console.print(operation_panel)

    def _rich_operation_details_panel(self, operation_name: str) -> Panel:

        operation_model = self.get_operation_model(operation_name)
        input_shape = get_input_shape(operation_model)
        output_shape = operation_model.output_shape

        input_shape_tree = Text("[yellow]No Input Shape Found")
        if input_shape:
            input_shape_tree = generate_rich_tree_from_shape(input_shape)
        output_shape_tree = generate_rich_tree_from_shape(output_shape)

        operation_docs = cleanhtml(operation_model.documentation)
        panel_group = Group(
            Padding(f"[bold underline]Documentation:[/] {operation_docs}", (1, 2)),
            Panel(
                input_shape_tree,
                title=f"Input: [yellow]{input_shape.name}",
                title_align="left",
                padding=(1, 1),
            ),
            Panel(
                output_shape_tree,
                title=f"Output: [yellow]{output_shape.name}",
                title_align="left",
                padding=(1, 1),
            ),
        )

        required_parameters = self.get_required_parameter_names_from_operation_name(
            operation_name
        )
        _title = f"[bold]Operation[/]: [magenta bold]{operation_name}[/]"
        if required_parameters:
            _title = f"{_title}   [white][Required: {', '.join(required_parameters)}]"
        operation_panel = Panel(
            panel_group, title=_title, highlight=True, title_align="left"
        )
        return operation_panel

    def find_best_relation_for_single_parameter(
        self, parameter_name: str, generated_relations_for_parameter: List[Dict]
    ) -> List[Dict]:
        if len(generated_relations_for_parameter) == 1:
            # found only one relation, no need to find the correct one
            return generated_relations_for_parameter

        # strip the required parameter name's identifiers
        _parameter_name_tokens = camel_case_split(parameter_name)
        non_id_parameter_tokens = [
            p_token
            for p_token in _parameter_name_tokens
            if p_token.lower() not in IDENTIFIER_NAMES
        ]

        # filter the gen. relation's resource name and required parameter name
        same_resource_name_relations = []
        for relation_obj in generated_relations_for_parameter:
            resource_node_name = relation_obj.resource_node_name
            resource_name_tokens = camel_case_split(resource_node_name)
            all_tokens_match = icompare_two_token_lists(
                non_id_parameter_tokens, resource_name_tokens
            )
            if all_tokens_match:
                same_resource_name_relations.append(relation_obj)

        if not same_resource_name_relations:
            return False

        # if there is only one relation with the same resource name, return it
        if len(same_resource_name_relations) == 1:
            return same_resource_name_relations

        # try to filter out the bad relations
        possible_relation_list = same_resource_name_relations
        relation_index_to_score_map = {
            str(i): 0 for i in range(len(possible_relation_list))
        }

        for i, relation_obj in enumerate(possible_relation_list):
            current_point = 0
            relations_operation_name = relation_obj.operation_name
            # promised to all have same operation
            # negative point if it has required parameters
            relations_required_parameters = (
                self.get_required_parameter_names_from_operation_name(
                    relations_operation_name
                )
            )
            current_point -= len(relations_required_parameters)

            # List verb gets +1 point
            if relations_operation_name.lower().startswith("list"):
                current_point += 1

            relation_index_to_score_map[str(i)] += current_point

        # select the best score
        selected_index_str, selected_score = max(
            relation_index_to_score_map.items(),
            key=lambda index_and_score: index_and_score[1],
        )
        selected_relation = possible_relation_list[int(selected_index_str)]

        if selected_relation:
            return [selected_relation]

        return False

    def get_pagination_token_output_to_parameter_name_mapping(
        self, operation_name: str
    ) -> Union[Dict[str, str], bool]:
        """Some Operations paginate their output using Pagination Token Parameters defined in `PAGINATION_TOKEN_KEYS`.

        Args:
            resource_node (ResourceNode): _description_
            operation_name (str): Name of the operation in the resource_node.

        Returns:
            Union[Dict[str, str], bool]: False or A dictionary with `parameter_name` and `output_key` keys. e.g. `{"parameter_name":"nextToken", "output_key": "NextToken"}`.
        """
        resource_node: ResourceNode = self

        # find all parameter names that are in `PAGINATION_TOKEN_KEYS`, case insensitive
        parameter_names = [
            pn
            for pn in resource_node.get_all_parameter_names_from_operation_name(
                operation_name
            )
            if is_word_in_a_list_of_words(pn, PAGINATION_TOKEN_KEYS)
        ]

        #  find all output keys that are in `PAGINATION_TOKEN_KEYS`, case insensitive
        output_keys = [
            ok
            for ok in resource_node.get_all_output_keys_from_operation_name(
                operation_name
            )
            if is_word_in_a_list_of_words(ok, PAGINATION_TOKEN_KEYS)
        ]

        if output_keys and parameter_names:
            return {
                "parameter_name": parameter_names[0],
                "output_key": output_keys[0],
            }
        return False

    def get_required_parameter_names_from_operation_model(self, operation_model):
        input_shape = get_input_shape(operation_model)
        if not input_shape:
            return []
        required_parameter_names = getattr(input_shape, "required_members", False)
        max_results_key = find_key_in_dict_keys("maxresults", required_parameter_names)
        if max_results_key:
            required_parameter_names.remove(max_results_key)
        return required_parameter_names

    def get_all_parameter_names_from_operation_name(self, operation_name):
        operation_model = self.get_operation_model(operation_name)
        input_shape = get_input_shape(operation_model)
        if not input_shape:
            return []

        all_parameter_names = []
        input_shape_and_target_paths = _flatten_shape_to_its_members_and_target_paths(
            input_shape
        )
        for shape_and_target_path in input_shape_and_target_paths:
            shape_name = get_shape_name(shape_and_target_path.shape)
            if shape_name:
                all_parameter_names.append(shape_name)
        return all_parameter_names

    def get_all_output_keys_from_operation_name(self, operation_name):
        operation_model = self.get_operation_model(operation_name)
        output_shape = operation_model.output_shape
        if not output_shape:
            return []

        all_parameter_names = []
        input_shape_and_target_paths = _flatten_shape_to_its_members_and_target_paths(
            output_shape
        )
        for shape_and_target_path in input_shape_and_target_paths:
            shape_name = get_shape_name(shape_and_target_path.shape)
            if shape_name:
                all_parameter_names.append(shape_name)
        return all_parameter_names

    # TODO: memoization
    def get_required_parameter_names_from_operation_name(
        self, operation_name: str
    ) -> List[str]:
        operation_model = self.get_operation_model(operation_name)
        return self.get_required_parameter_names_from_operation_model(operation_model)

    def get_all_required_parameter_names(self) -> List[str]:
        all_required_names = []
        for operation_name in self.operation_names:
            required_shapes = self.get_required_parameter_names_from_operation_name(
                operation_name
            )
            all_required_names.extend(required_shapes)
        return all_required_names

    def get_operation_model(self, operation_name: str) -> OperationModel:
        if operation_name in self._operation_models:
            return self._operation_models[operation_name]
        service_model = self.service_node.get_service_model()
        operation_model = service_model.operation_model(operation_name)
        self._operation_models[operation_name] = operation_model
        return operation_model

    def json(self) -> Dict:
        return {
            "service_node_name": self.service_node.name,
            "name": self.name,
            "operation_names": self.operation_names,
        }

    def __str__(self) -> str:
        # return f"[{self.service_node.name}.{self.name}]"
        return f"[{self.name}]"


class YamlResourceNode(ResourceNode):
    def __init__(
        self,
        service_node: "ServiceNode",
        name: str,
        operation_names: List[str],
        yaml_config: Dict = None,
    ) -> None:
        super().__init__(service_node, name, operation_names)
        self.yaml_config = yaml_config

    def define_extra_relations(self) -> Union[List[Dict], List[Relation]]:
        if not self.yaml_config.extra_relations:
            return []
        extra_relations = [
            Relation(**extra_relation.__dict__)
            for extra_relation in self.yaml_config.extra_relations
        ]
        return extra_relations

    # NOTE: +overrideable
    def get_operations_relations(
        self, operation_name: str
    ) -> Tuple[Union[List[Dict], bool], Union[Error, None]]:
        operations = self.yaml_config.operations
        for operation in operations:
            if operation_name == operation.operation_name:
                # cast schema to Relation objects
                if operation.explicit_relations:
                    explicit_relations = [
                        Relation(**rel_model.__dict__)
                        for rel_model in operation.explicit_relations
                    ]
                    return explicit_relations, None
        # if no explicit relations are defined, then call the ResourceNode method and return it
        return super().get_operations_relations(operation_name)

    # NOTE: +overrideable
    def generate_jmespath_selector_from_relations(
        self, operation_name: str, relation_list: List[Dict]
    ) -> str:
        operations = self.yaml_config.operations
        for operation in operations:
            if operation_name == operation.operation_name:
                if operation.jmespath_selector:
                    js_selector = operation.jmespath_selector
                    return js_selector

        return super().generate_jmespath_selector_from_relations(
            operation_name, relation_list
        )

    # NOTE: +overrideable
    def complement_api_parameters_list(
        self,
        operation_name: str,
        related_operations_data: Union[List, Dict],
        relations_of_operation: List[Dict],
        raw_api_parameters_list: List,
    ) -> List:

        precomplemented_api_params = super().complement_api_parameters_list(
            operation_name,
            related_operations_data,
            relations_of_operation,
            raw_api_parameters_list,
        )

        for operation in self.yaml_config.operations:
            if operation_name == operation.operation_name:
                if operation.complement_api_parameters:
                    # if the option is defined
                    for complement_action in operation.complement_api_parameters:
                        if complement_action.action == "add":
                            data = complement_action.data
                            for api_param_item in precomplemented_api_params:
                                api_param_item.update(data)

                        elif complement_action.action == "remove":
                            for api_param_item in precomplemented_api_params:
                                for remove_key in complement_action.keys:
                                    api_param_item.pop(remove_key, None)

        return precomplemented_api_params

    def generate_api_parameters_from_operation_data(
        self,
        operation_name: str,
        relations_of_operation: List[Dict],
        related_operations_data: Union[List, Dict],
    ) -> Tuple[Union[List, bool], Union[Error, None]]:
        """
        get a list for api parameters, override them
        """
        operations = self.yaml_config.operations
        for operation in operations:
            if operation_name == operation.operation_name:
                override_parameters = operation.override_api_parameters
                if override_parameters:
                    return override_parameters, None

        # if not defined, fall back to the default super class method
        return super().generate_api_parameters_from_operation_data(
            operation_name, relations_of_operation, related_operations_data
        )


class ServiceNode:
    def __init__(self, name, session):
        self.name = name
        self.session = session
        self.client = self.session.client(self.name)
        self.resource_nodes = None
        self._relation_map = None
        self._reader = None
        self._read_operation_name_to_tokens_map = None

    def get_client(self):
        return self.client

    def read(self, resource_node_name):
        reader = self.get_service_reader()
        return reader.read_resource_node(resource_node_name)

    def get_service_reader(self) -> ServiceReader:
        """Returns/creates the ServiceReader for the current ServiceNode

        Returns:
            ServiceReader: ServiceReader object for current ServiceNode
        """
        if not self._reader:
            self._reader = ServiceReader(self)
        return self._reader

    def print_resource_node(self, resource_node_name: str) -> None:
        # TODO: check if resource node exists
        operations_panel = self._get_operation_details_panel(resource_node_name)
        if operations_panel:
            console.print(operations_panel)

    def _get_operation_details_panel(self, resource_node_name: str) -> Panel:
        resource_node = self.get_resource_node_by_name(resource_node_name)
        if not resource_node:
            return False
        panels_for_operations = [
            resource_node._rich_operation_details_panel(operation_name)
            for operation_name in resource_node.operation_names
        ]

        operations_group = Group(*panels_for_operations)
        return Panel(
            operations_group,
            title=f"[bold]Resource Node: [blue]{resource_node.name} ",
            title_align="left",
        )

    def get_resource_node_by_name(self, resource_node_name: str) -> ResourceNode:
        """Searches the current ServiceNode for the given `resource_node_name`,
        and returns it.

        Args:
            resource_node_name (str): Name of the ResourceNode

        Returns:
            ResourceNode: The ResourceNode object, `False` if not found.
        """
        if not resource_node_name:
            return False

        for r_node in self.get_resource_nodes():
            # FIXME: laterr
            # if r_node.name == resource_node_name:
            if icompare_two_camel_case_words(r_node.name, resource_node_name):
                return r_node
        return False

    def json(self) -> Dict:
        return {
            "service_name": self.name,
            # TODO:
        }

    # TODO: memoization
    def find_resource_node_by_operation_name(self, operation_name: str) -> ResourceNode:
        """Traverses the ResourceNodes of the current ServiceNode and tries to find
        the ResourceNode that has the `operation_name` in it.

        Args:
            operation_name (str): Name of the operation

        Returns:
            ResourceNode: ResourceNode object that has the `operation_name`, or None.
        """
        for r_node in self.get_resource_nodes():
            if operation_name in r_node.operation_names:
                return r_node
        return None

    def get_relation_map(self) -> RelationMap:
        """Gets the relation map object.

        Returns:
            RelationMap: RelationMap object for the current ServiceNode
        """
        if self._relation_map is not None:
            return self._relation_map
        self._relation_map = RelationMap(self)
        return self._relation_map

    def get_resource_nodes(self) -> List[ResourceNode]:
        """Gets the available `ResourceNode`s of the current ServiceNode

        Returns:
            List[ResourceNode]: List of `ResourceNode`s available in the ServiceNode
        """
        if self.resource_nodes is None:
            self.resource_nodes = self._generate_resource_nodes()
        return self.resource_nodes

    def create_resource_node(self, **kwargs: Dict) -> ResourceNode:
        """Creates the ResourceNode object with the given `kwargs`.
        Uses the `ResourceNodeRegistry` to find the custom subclasses of the
        `ResourceNode` class, else defaults to use the ResourceNode class

        Raises: # TODO: fix this, return false instead
            Exception: _description_

        Returns:
            ResourceNode: _description_
        """
        service_name = self.name
        resource_node_name = kwargs.get("name", False)
        _ResourceNodeClass = ResourceNode
        if not resource_node_name:
            logger.debug("A resource name must be provided on creation.", extra=kwargs)
            raise Exception("ResourceNode name must be provided.")
        _custom_cls_for_resource_node = (
            _resource_node_registry.find_custom_class_for_resource_node(
                service_name, resource_node_name
            )
        )
        if _custom_cls_for_resource_node:
            logger.debug(
                f"ResourceNodeRegistry: [bold][green]{service_name}[/].[blue]{resource_node_name}[/][/] has extended with: {_custom_cls_for_resource_node.__name__}"
            )
            _ResourceNodeClass = _custom_cls_for_resource_node
            return _ResourceNodeClass(**kwargs)

        # try to find yaml config defined for this ResourceNode
        yaml_config_found = _resource_node_registry.search_yaml_config_registry(
            service_name, resource_node_name
        )
        if yaml_config_found:
            # add the yaml_config for the YamlResourceNode subclass initialization
            kwargs.update({"yaml_config": yaml_config_found})
            yaml_resource_node_obj = YamlResourceNode(**kwargs)
            return yaml_resource_node_obj

        # return default ResourceNode class
        return _ResourceNodeClass(**kwargs)

    def _generate_resource_nodes(self) -> List[ResourceNode]:
        """Parses `botocore` client for the AWS service and generates ResourceNodes.
        Generated `ResourceNode`s can be subclasses of `ResourceNode`.

        Returns:
            List[ResourceNode]: List of generated `ResourceNode`s
        """
        generated_resouce_nodes = []
        op_name_to_tokens_map = self.get_read_operation_name_to_tokens_map()
        resource_name_to_operations_map = {}
        for operation_name, op_name_tokens in op_name_to_tokens_map.items():
            verb, *resource_name_list = op_name_tokens
            resource_node_name = "".join(resource_name_list)

            found_key = ifind_key_in_dict_keys(
                resource_node_name, resource_name_to_operations_map.keys()
            )
            if found_key:
                # decide if we want to keep the plural or singular name

                sp_comparison = compare_nouns(resource_node_name, found_key)
                if sp_comparison == "s:p":
                    # swap the found_key with resource_node_name as dict key
                    resource_name_to_operations_map[resource_node_name] = [
                        operation_name
                    ] + resource_name_to_operations_map[found_key]
                    resource_name_to_operations_map.pop(found_key)
                else:
                    resource_name_to_operations_map[found_key].append(operation_name)
            else:
                resource_name_to_operations_map[resource_node_name] = [operation_name]

        # createt the resource nodes
        for (
            resource_node_name,
            operations_list,
        ) in resource_name_to_operations_map.items():
            created_resource_node = self.create_resource_node(
                service_node=self,
                name=resource_node_name,
                operation_names=operations_list,
            )
            if created_resource_node:
                generated_resouce_nodes.append(created_resource_node)

        if len(generated_resouce_nodes) <= 1:
            return generated_resouce_nodes

        # TODO: decide if this should be kept
        # concat same names
        combined_resource_nodes = []
        _used_for_combining_resource_nodes = []

        for i, gen_resource_node in enumerate(generated_resouce_nodes[:-1]):
            if gen_resource_node in _used_for_combining_resource_nodes:
                continue
            for other_resource_node in generated_resouce_nodes[i + 1 :]:
                # FIXME: IMPORTANT
                if compare_two_camel_case_words(
                    gen_resource_node.name, other_resource_node.name
                ):
                    # select the shortest name, meaning singular
                    singular_named_resource_node = gen_resource_node
                    plural_named_resource_node = other_resource_node
                    if len(singular_named_resource_node.name) > len(
                        plural_named_resource_node.name
                    ):
                        singular_named_resource_node, plural_named_resource_node = (
                            plural_named_resource_node,
                            singular_named_resource_node,
                        )
                    _used_for_combining_resource_nodes.append(
                        plural_named_resource_node
                    )
                    singular_named_resource_node.operation_names.extend(
                        plural_named_resource_node.operation_names
                    )

            if gen_resource_node not in _used_for_combining_resource_nodes:
                combined_resource_nodes.append(gen_resource_node)

        return combined_resource_nodes

    def get_read_operation_name_to_tokens_map(self) -> Dict:
        """Generate `operation name to word tokens` map for the
        available read operations in the ServiceNode.

        Caches the output in the class attr. `_read_operation_name_to_tokens_map`.

        Returns:
            Dict: dictionary
        """
        if self._read_operation_name_to_tokens_map is not None:
            return self._read_operation_name_to_tokens_map
        read_operation_names = self.get_read_operation_names()
        self._read_operation_name_to_tokens_map = (
            self._generate_operation_name_to_tokens_map(read_operation_names)
        )
        return self._read_operation_name_to_tokens_map

    def _generate_operation_name_to_tokens_map(
        self, operation_names: List[str] = None
    ) -> Dict:
        if operation_names is None:
            operation_names = self.get_operation_names()

        op_name_to_tokens = {
            op_name: camel_case_split(op_name) for op_name in operation_names
        }
        return op_name_to_tokens

    def get_read_operation_names(self) -> List[str]:
        """Gets the operation names from the boto3 client and
        filters the operation names starting with `List`,`Get` or `Describe`.

        Returns:
            List[str]: Read Only operation names
        """
        operation_names = self.get_operation_names()
        read_only_operation_names = [
            op_name
            for op_name in operation_names
            if any([op_name.startswith(read_verb) for read_verb in READ_ONLY_VERBS])
        ]
        return read_only_operation_names

    def get_service_model(self) -> ServiceModel:
        """Returns the ServiceModel obj from boto3 client"""
        service_model = self.client._service_model
        return service_model

    def get_event_system(self) -> EventAliaser:
        """Returns the boto3 clients event system"""
        event_system = self.client.meta.events
        return event_system

    def get_operation_names(self) -> List[str]:
        """Returns the `operation_names` defined in the boto client._service_model"""
        service_model = self.get_service_model()
        operation_names = service_model.operation_names
        return operation_names



from typing import List, Dict, Optional, Any
from pydantic import BaseModel, validator
import yaml


class YamlRelation(BaseModel):
    service_name: str
    resource_node_name: str
    operation_name: str
    required_shape_name: str
    target_shape_name: str
    target_shape_type: str
    target_path: str


class YamlComplementApiParameterAction(BaseModel):
    _ACTION_TYPES = ("add", "remove")

    action: str
    data: Optional[Dict[str, Any]]
    keys: Optional[List[str]]

    @validator("action")
    def action_must_be_valid(cls, v):
        if v not in cls._ACTION_TYPES:
            raise ValueError("action must be 'add' or 'remove'")
        return v


class YamlResourceNodeOperation(BaseModel):
    operation_name: str
    jmespath_selector: Optional[str]
    complement_api_parameters: Optional[List[YamlComplementApiParameterAction]]
    explicit_relations: Optional[List[YamlRelation]]
    override_api_parameters: Optional[List[Dict[str, Any]]]


class YamlServiceResourceNode(BaseModel):
    resource_node_name: str
    extra_relations: Optional[List[YamlRelation]]
    operations: Optional[List[YamlResourceNodeOperation]]


class YamlService(BaseModel):
    service_name: str
    resource_nodes: Optional[List[YamlServiceResourceNode]]

    # @validator("resource_node")
    # def validate_resource_node(cls, value):
    #     if not value.extra_relations and not value.operations:
    #         raise ValueError(
    #             "Either extra_relations or operations must be defined in resource_node"
    #         )
    #     return value


# Load input data from YAML
def test_service_yaml():
    input_yaml = """
    service_name: ec2
    resource_nodes:
    - resource_node_name: Instance
    extra_relations:
        - service_name: x
        resource_node_name: y 
        operation_name: z
        required_shape_name: a 
        target_shape_name: b
        target_shape_type: c
        target_path: d
    operations:
        - operation_name: ListOperation
        jmespath_selector: "[].example"
        complement_api_parameters:
            - action: add
            data:
                any: data
                is: fine
            - action: remove
            keys:
                - aa
                - bb
        explicit_relations:
            - service_name: x
            resource_node_name: y
            operation_name: z
            required_shape_name: a
            target_shape_name: b
            target_shape_type: c
            target_path: d
            - service_name: x1
            resource_node_name: y1
            operation_name: z1
            required_shape_name: a1
            target_shape_name: b1
            target_shape_type: c1
            target_path: d1
        override_api_parameters:
            - param1: arg1
            param2: param2
            param3: arg3
    """
    input_data = yaml.safe_load(input_yaml)

    # Validate input data using the Service model
    service = YamlService(**input_data)
    print(service)



from config import get_logger
from yaml_validators import YamlService

import yaml
import os
from typing import List, Dict, Union, Tuple
from pathlib import Path

YAML_IGNORE_PREFIX = "_"
YAML_SERVICES_DIRECTORY = Path(__file__).parent / "custom_nodes" / "yamls"

logger = get_logger(__name__)


def parse_yaml_file_to_service(
    yaml_file_path: str,
) -> Tuple[YamlService, Union[Exception, None]]:
    """Parses given YAML file and returns a pydantic model: YamlService object

    Args:
        yaml_file_path (str): Filepath for the yaml file

    Returns:
        YamlService: Input data parsed as a pydantic YamlService object
        Exception or `None`: Exception if raised during parsing, or None if successful
    """
    try:
        with open(yaml_file_path, "r") as f:
            input_data = yaml.load(f, Loader=yaml.FullLoader)
        return YamlService(**input_data), None
    except Exception as e:
        return False, e


def find_all_yaml_files(directory: str) -> List[str]:
    """Find all yaml files in a directory with the exception of files starting with "_".

    Args:
        directory (str): Directory to search for yaml files

    Returns:
        List[str]: List of yaml file paths
    """
    yaml_files = []
    for root, _, files in os.walk(directory):
        for file in files:
            # support any yaml file with the exception of files starting with "YAML_IGNORE_PREFIX"
            if file.endswith(".yaml") and not file.startswith(YAML_IGNORE_PREFIX):
                yaml_files.append(os.path.join(root, file))
    return yaml_files


def find_and_parse_yaml_services() -> List[YamlService]:
    """Searches and finds the defined yaml files in the YAML_SERVICES_DIRECTORY directory that are not starting with '_'

    Returns:
        List[YamlService]: List of found YamlServices, created from yaml files
    """
    found_yaml_services = []
    yaml_files = find_all_yaml_files(YAML_SERVICES_DIRECTORY)
    logger.debug(
        f"Found {len(yaml_files)} yaml files in {YAML_SERVICES_DIRECTORY}. Starting to parse & validate them."
    )
    for yaml_file in yaml_files:
        yaml_service, error = parse_yaml_file_to_service(yaml_file)
        if error is None and yaml_service:
            found_yaml_services.append(yaml_service)
        else:
            logger.debug(
                f"Failed to parse [bold]{yaml_file}[/] to service. Error: {str(error)}. "
            )

    return found_yaml_services



from aws import BalconyAWS
from utils import is_word_in_a_list_of_words
from typing import Dict, List, Union
from nodes import PAGINATION_TOKEN_KEYS

def get_pagination_token_output_to_parameter_name_mapping(
    self, operation_name: str
) -> Union[Dict[str, str], bool]:
    """Some Operations paginate their output using Pagination Token Parameters defined in `PAGINATION_TOKEN_KEYS`.

    Args:
        resource_node (ResourceNode): _description_
        operation_name (str): Name of the operation in the resource_node.

    Returns:
        Union[Dict[str, str], bool]: False or A dictionary with `parameter_name` and `output_key` keys. e.g. `{"parameter_name":"nextToken", "output_key": "NextToken"}`.
    """
    resource_node = self

    # find all parameter names that are in `PAGINATION_TOKEN_KEYS`, case insensitive
    parameter_names = [
        pn
        for pn in resource_node.get_all_parameter_names_from_operation_name(
            operation_name
        )
        if is_word_in_a_list_of_words(pn, PAGINATION_TOKEN_KEYS)
    ]

    #  find all output keys that are in `PAGINATION_TOKEN_KEYS`, case insensitive
    output_keys = [
        ok
        for ok in resource_node.get_all_output_keys_from_operation_name(
            operation_name
        )
        if is_word_in_a_list_of_words(ok, PAGINATION_TOKEN_KEYS)
    ]

    if output_keys and parameter_names:
        return {
            "parameter_name": parameter_names[0],
            "output_key": output_keys[0],
        }
    return False


def main():
    baws = BalconyAWS()
    aws_services = baws.get_available_service_names()
    for aws_service_name in aws_services[aws_services.index("lexv2-runtime") :]:
        print(aws_service_name)
        service_node = baws.get_service_node(aws_service_name)
        for resource_node in service_node.get_resource_nodes():
            for operation_name in resource_node.get_operation_names():
                a = get_pagination_token_output_to_parameter_name_mapping(
                    resource_node, operation_name
                )
                print(a)

                # "nexttoken, continuationtoken, marker"
                # rpp = [_ for _ in rp if "token" in _.lower() or "marker" in _.lower()]
                # opp = [_ for _ in op if "token" in _.lower() or "marker" in _.lower()]
                # if rpp or opp:
                #     print(
                #         f"  {operation_name} :: {','.join(rpp)} --::-- {','.join(opp)}"
                #     )


def __():
    pass


main()



from nodes import ResourceNode
from config import get_logger
from relations import Relation
import jmespath

logger = get_logger(__name__)


class RolePolicy(ResourceNode, service_name="iam", name="RolePolicy"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_operations_relations(self, operation_name: str):
        """
        Relations defined in this function overrides the auto generated relations.
        Operations given in the relations are promised to be called before the current operation.
        And the related operations data will be passed to the generate_api_parameters_from_operation_data function.

        `GetRolePolicy` operations parameters can be generated from the `ListRolePolicies` operations output.
        """
        if operation_name == "GetRolePolicy":
            return [
                Relation(
                    **{
                        "service_name": "iam",
                        "resource_node_name": "RolePolicy",
                        "operation_name": "ListRolePolicies",
                        "required_shape_name": "PolicyName",
                        "target_shape_name": "PolicyNames",
                        "target_shape_type": "list",
                        "target_path": "--ommitted--",
                    }
                )
            ], None

        return super().get_operations_relations(operation_name)

    def generate_api_parameters_from_operation_data(
        self, operation_name, relations_of_operation, related_operations_data
    ):
        """
        GetRolePolicy requires RoleName and PolicyName as parameters which can be found in the ListRolePolicies output.
        The reason to override this method is to generate the valid required parameters which was not possible with only the jmespath query.

        ListRolePolicies output data looks like this:
        ```
        [
            {
                "PolicyNames": [
                    "_policy_name_1",
                    "_policy_name_2"
                ],
                "IsTruncated": false,
                "__args__": {
                    "RoleName": "_role_name"
                }
            },
        ]
        ```

        and GetRolePolicy operation requires {"RoleName":..., "PolicyName":...} as parameters.
        This function generates the required parameters from the ListRolePolicies output if the operation is GetRolePolicy.
        """
        if operation_name == "GetRolePolicy":
            generated_api_parameters = []

            # select non-empty PolicyNames lists
            non_empty_policy_names = jmespath.search(
                "ListRolePolicies[?length(PolicyNames) > `0`]", related_operations_data
            )
            for data in non_empty_policy_names:
                role_name = data.get("__args__").get("RoleName")
                policy_names = data.get("PolicyNames")
                for policy_name in policy_names:
                    generated_api_parameters.append(
                        {"RoleName": role_name, "PolicyName": policy_name}
                    )
            return generated_api_parameters, None

        # if this is NOT our selected operation, let the normal flow run
        return super().generate_api_parameters_from_operation_data(
            operation_name, relations_of_operation, related_operations_data
        )


class UserPolicy(ResourceNode, service_name="iam", name="UserPolicy"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_operations_relations(self, operation_name: str):

        if operation_name == "GetUserPolicy":
            return [
                Relation(
                    **{
                        "service_name": "iam",
                        "resource_node_name": "UserPolicy",
                        "operation_name": "ListUserPolicies",
                        "required_shape_name": "--ommitted--",
                        "target_shape_name": "--ommitted--",
                        "target_shape_type": "--ommitted--",
                        "target_path": "--ommitted--",
                    }
                )
            ], None

        return super().get_operations_relations(operation_name)

    def generate_api_parameters_from_operation_data(
        self, operation_name, relations_of_operation, related_operations_data
    ):
        if operation_name == "GetUserPolicy":
            generated_api_parameters = []

            # select non-empty PolicyNames lists
            non_empty_policy_names = jmespath.search(
                "ListUserPolicies[?length(PolicyNames) > `0`]", related_operations_data
            )
            for data in non_empty_policy_names:
                user_name = data.get("__args__").get("UserName")
                policy_names = data.get("PolicyNames")
                for policy_name in policy_names:
                    generated_api_parameters.append(
                        {"UserName": user_name, "PolicyName": policy_name}
                    )
            return generated_api_parameters, None

        # if this is NOT our selected operation, let the normal flow run
        return super().generate_api_parameters_from_operation_data(
            operation_name, relations_of_operation, related_operations_data
        )


class GroupPolicy(ResourceNode, service_name="iam", name="GroupPolicy"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_operations_relations(self, operation_name: str):

        if operation_name == "GetGroupPolicy":
            return [
                Relation(
                    **{
                        "service_name": "iam",
                        "resource_node_name": "GroupPolicy",
                        "operation_name": "ListGroupPolicies",
                        "required_shape_name": "--ommitted--",
                        "target_shape_name": "--ommitted--",
                        "target_shape_type": "--ommitted--",
                        "target_path": "--ommitted--",
                    }
                )
            ], None

        return super().get_operations_relations(operation_name)

    def generate_api_parameters_from_operation_data(
        self, operation_name, relations_of_operation, related_operations_data
    ):
        if operation_name == "GetGroupPolicy":
            generated_api_parameters = []

            # select non-empty PolicyNames lists
            non_empty_policy_names = jmespath.search(
                "ListGroupPolicies[?length(PolicyNames) > `0`]", related_operations_data
            )
            for data in non_empty_policy_names:
                group_name = data.get("__args__").get("GroupName")
                policy_names = data.get("PolicyNames")
                for policy_name in policy_names:
                    generated_api_parameters.append(
                        {"GroupName": group_name, "PolicyName": policy_name}
                    )
            return generated_api_parameters, None

        # if this is NOT our selected operation, let the normal flow run
        return super().generate_api_parameters_from_operation_data(
            operation_name, relations_of_operation, related_operations_data
        )



from utils import get_all_available_services, ifind_similar_names_in_list
from config import (
    get_logger,
    get_rich_console,
    set_log_level_at_runtime,
    clear_relations_cache,
)
from custom_nodes import *  # noqa
from aws import BalconyAWS

import typer
import jmespath
from typing import Optional, List
from rich.columns import Columns
from rich.panel import Panel
import logging
import boto3

console = get_rich_console()
logger = get_logger(__name__)
session = boto3.session.Session()
service_factory = BalconyAWS(session)
app = typer.Typer(no_args_is_help=True)


@app.callback()
def _main_app_callback(
    debug: bool = typer.Option(False, "--debug", "-d", help="Enable debug messages."),
):
    if debug:
        set_log_level_at_runtime(logging.DEBUG)


def _get_available_service_node_names():
    return get_all_available_services(session)


def generate_service_node_completion_items():
    service_names = _get_available_service_node_names(session)
    return service_names


def _complete_service_name(incomplete: str):
    service_names = _get_available_service_node_names()
    if not incomplete:
        for name in service_names:
            yield name
    else:
        for name in service_names:
            if name.startswith(incomplete):
                yield name


def _complete_resource_node_name(ctx: typer.Context, incomplete: str) -> List[str]:
    service = ctx.params.get("service", False)
    if not service:
        return []
    service_node = service_factory.get_service_node(service)
    if not service_node:
        return []
    resource_nodes = service_node.get_resource_nodes()
    if not resource_nodes:
        return []

    resource_node_names = []

    for _rn in resource_nodes:
        _rn_name = _rn.name
        if incomplete:
            if _rn_name.lower().startswith(incomplete.lower()):
                resource_node_names.append(_rn_name)
        else:
            resource_node_names.append(_rn_name)
    return resource_node_names


def _complete_operation_type(ctx: typer.Context) -> List[str]:
    service = ctx.params.get("service", False)
    resource_node = ctx.params.get("resource_node", False)
    if not service or not resource_node:
        return []
    service_node = service_factory.get_service_node(service)
    if not service_node:
        return []
    resource_node_obj = service_node.get_resource_node_by_name(resource_node)
    if not resource_node_obj:
        return []
    operation_types_and_names = resource_node_obj.get_operation_types_and_names()
    return list(operation_types_and_names.keys())


def _list_service_or_resource(
    service: Optional[str] = typer.Argument(
        None,
        show_default=False,
        help="The AWS service name",
        autocompletion=_complete_service_name,
    ),
    resource_node: Optional[str] = typer.Argument(
        None,
        show_default=False,
        help="The AWS Resource Node",
        autocompletion=_complete_resource_node_name,
    ),
) -> None:

    available_service_names = _get_available_service_node_names()
    if not service and not resource_node:
        # nothing is given
        console.print(Columns(available_service_names, equal=True, expand=True))
        return available_service_names

    if service and not resource_node:
        # we only have service name

        if service not in available_service_names:
            similar_service_names = ifind_similar_names_in_list(
                service, available_service_names
            )
            if similar_service_names:
                console.print("[yellow]Found Similar Service Names:")
                console.print(Columns(similar_service_names, equal=True, expand=True))
                return
            else:
                raise typer.Exit(
                    f"Invalid service name: {service}. Please pick a proper one."
                )

        service_node = service_factory.get_service_node(service)
        resource_nodes = service_node.get_resource_nodes()

        resource_node_names = []
        for _rn in resource_nodes:
            _rn_name = _rn.name
            if len(_rn.get_all_required_parameter_names()) >= 2:
                _rn_name = f"[bold red]{_rn_name}[/]"
            elif len(_rn.get_all_required_parameter_names()) >= 1:
                _rn_name = f"[bold green]{_rn_name}[/]"
            resource_node_names.append(_rn_name)

        resource_node_name_as_columns = Columns(
            resource_node_names, equal=True, expand=True
        )
        console.print(resource_node_name_as_columns)
        return resource_node_names

    elif service and resource_node:
        # we got both options filled
        service_node = service_factory.get_service_node(service)
        resource_nodes = service_node.get_resource_nodes()
        resource_node_names = [_rn.name for _rn in resource_nodes]
        resource_node_obj = service_node.get_resource_node_by_name(resource_node)
        if not resource_node_obj:

            similar_resource_names = ifind_similar_names_in_list(
                resource_node, resource_node_names
            )
            if similar_resource_names:
                console.print("[yellow]Found Similar Resource Node Names:")
                console.print(Columns(similar_resource_names, equal=True, expand=True))
                return
            else:
                raise typer.Exit(
                    f"Invalid Resource Node name: [bold]{resource_node}[/] for Service: [bold]{service}[/]."
                )
        operations_panel = service_node._get_operation_details_panel(
            resource_node_obj.name
        )
        console.print(operations_panel)
        return


@app.command("aws", help="List AWS services, Call read-operations, Show documentation")
def aws_main_command(  # noqa
    service: Optional[str] = typer.Argument(
        None,
        show_default=False,
        help="Name of the AWS Service",
        autocompletion=_complete_service_name,
    ),
    resource_node: Optional[str] = typer.Argument(
        None,
        show_default=False,
        help="Name of the AWS Resource Node",
        autocompletion=_complete_resource_node_name,
    ),
    operation: Optional[str] = typer.Argument(
        None,
        show_default=False,
        help="Select a specific operation type. <get, gets, list, describe>",
        autocompletion=_complete_operation_type,
    ),
    patterns: Optional[List[str]] = typer.Option(
        None,
        "--pattern",
        "-p",
        show_default=False,
        help='UNIX pattern matching for generated parameters. Should be quoted. e.g. (-p "*prod-*")',
    ),
    jmespath_selector: Optional[str] = typer.Option(
        None,
        "--jmespath-selector",
        "-js",
        show_default=False,
        help="JMESPath query selector to filter resulted data. Visit for tutorial: https://jmespath.org/tutorial.html",
    ),
    debug: bool = typer.Option(False, "--debug", "-d", help="Enable debug messages."),
    formatter: Optional[str] = typer.Option(
        None,
        "--format",
        "-f",
        show_default=False,
        help="Python f-string expression to generate a line for each item.",
    ),
    list_contents: bool = typer.Option(
        False,
        "--list",
        "-l",
        help="Print the details of Service or Resource. Does not make requests.",
    ),
    paginate: bool = typer.Option(
        False, "--screen", "-s", help="Open the data on a separate paginator on shell."
    ),
):
    if debug:
        set_log_level_at_runtime(logging.DEBUG)
    if list_contents:
        _list_service_or_resource(service, resource_node)
        return

    if not service and not resource_node:
        available_services = _list_service_or_resource(service, resource_node)
        console.print(
            Panel("[bold]Please pick one of the AWS Services", title="[red][bold]ERROR")
        )
        return {"services": available_services}

    if service and not resource_node:
        available_resources = _list_service_or_resource(service, resource_node)
        console.print(
            Panel(
                f"[bold]Please pick one of the Resource Nodes from [green]{service}[/] Service",
                title="[red][bold]ERROR",
            )
        )
        return {"service": service, "resources": available_resources}

    elif service and resource_node:
        service_node = service_factory.get_service_node(service)
        service_reader = service_node.get_service_reader()

        is_operation_selected = operation is not None
        read_data = None
        if not is_operation_selected:
            # read all operations in given resource node
            read_data = service_reader.read_resource_node(
                resource_node, match_patterns=patterns
            )

        else:  # Operation is selected
            resource_node_obj = service_node.get_resource_node_by_name(resource_node)
            types_to_op_names = resource_node_obj.get_operation_types_and_names()
            supported_operation_types = list(types_to_op_names.keys())
            operation_name = types_to_op_names.get(operation, False)
            if not operation_name:
                console.print(
                    f"[red bold]Given {operation} is not supported by {resource_node}. Try: {supported_operation_types}"
                )
                return False
            read_data = service_reader.read_operation(
                resource_node, operation_name, match_patterns=patterns
            )

        if jmespath_selector:

            logger.debug(
                f"Using jmespath selector: {jmespath_selector} to query the returned data."
            )
            read_data = jmespath.search(jmespath_selector, read_data)

        if formatter:
            if read_data:
                for r_data in read_data:
                    try:
                        console.print(formatter.format(**r_data), overflow="ignore")
                    except AttributeError as e:
                        logger.debug(
                            f"Failed to format  [red]{str(e)}[/] with data: {r_data}"
                        )
                return read_data
            else:
                logger.debug(
                    f"[red]No data found.[/] Failed to use --format: '{formatter}'"
                )

        if paginate:
            with console.pager(styles=True):
                console.print_json(data=read_data, default=str)
        else:
            console.print_json(data=read_data, default=str)
        return read_data


@app.command("clear-cache", help="Clear relations json cache")
def clear_cache_command(
    # service: Optional[str] = typer.Argument(None, show_default='all',
    # help='Name of the Service to clear relation caches of', autocompletion=_complete_service_name),
):
    logger.debug("Deleting relations cache for services:")
    deleted_service_caches = clear_relations_cache()
    for deleted_service in deleted_service_caches:
        logger.info(f"[green]Deleted[/] {deleted_service}")


# @app.command('version', help='Show version info' )
# def version_command():
#     from importlib.metadata import version
#     console.print(__package__)
#     version=version(__package__)
#     console.print(version)


def run_app():
    app(prog_name="balcony")


if __name__ == "__main__":
    run_app()



"""
Any resource that app system is going to use must be exported here.
Otherwise different import statements would create new global static objects
for AppRegistry etc.
"""
import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

try:
    from .aws import BalconyAWS
    from .config import get_rich_console, get_logger
    from .nodes import ServiceNode, ResourceNode
    from .reader import ServiceReader
    from .relations import RelationMap
    from .errors import Error
except ImportError:
    # print('ImportError: balcony/__init__.py')
    from aws import BalconyAWS
    from config import get_rich_console, get_logger
    from nodes import ServiceNode, ResourceNode
    from reader import ServiceReader
    from relations import RelationMap
    from errors import Error



from nodes import ResourceNode
from config import get_logger

logger = get_logger(__name__)

class BucketLifecycleConfiguration(ResourceNode, service_name="s3", name="BucketLifecycleConfiguration"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def define_extra_relations(self):
        r= super().define_extra_relations()
        return [{
            "service_name": "s3",
            "resource_node_name": "Buckets",
            "required_shape_name": "Bucket",
            "target_shape_name": "Name",
            "target_shape_type": "string",
            "operation_name": "ListBuckets",
            "target_path": "Buckets[*].Name"
        }]

    # def find_best_relations_for_operation(self, operation_name, relation_map):
    #     r = super().find_best_relations_for_operation(operation_name, relation_map)
    #     return r
    
    def generate_jmespath_selector_from_relations(self, operation_name, relation_list):
        r = super().generate_jmespath_selector_from_relations(operation_name, relation_list)
        return r



from utils import get_all_available_services, _create_boto_session
from nodes import ServiceNode
from reader import ServiceReader

from typing import Optional, List, Union
import boto3


# class Boto3SessionSingleton(object):
#     _instance = None
#     _session = None

#     def __new__(cls):
#         if cls._instance is None:
#             cls._instance = super(Boto3SessionSingleton, cls).__new__(cls)
#             # Put any initialization here.
#             # cls._session = _create_boto_session()
#         return cls._instance

#     def __init__(self):
#         if not self._session:
#             self._session = _create_boto_session()

#     def get_session(self):
#         return self._session


class BalconyAWS:
    """Represents the whole AWS API.
    Can be used to get ServiceNodes and ServiceReaders.
    Operations can be read with `read_` functions.

    ```python title="Creating a BalconyAWS obj and reading IAM Roles"
    baws = BalconyAWS()
    roles = baws.read_resource_node('iam', 'Role')
    print(roles)
    ```
    """

    def __init__(self, boto3_session: Optional[boto3.session.Session] = None):
        self.boto3_session = boto3_session
        if boto3_session is None:
            self.boto3_session = _create_boto_session()
        self._service_nodes_map = {}

    def _create_service_node(self, service_name: str) -> None:
        """Creates the ServiceNode w/ the `self.boto3_session`

        Args:
            service_name (str): Name of the AWS Service
        """
        service_node = ServiceNode(service_name, self.boto3_session)
        self._service_nodes_map[service_name] = service_node

    def get_service_node(self, service_name: str) -> ServiceNode:
        """Gets or creates the ServiceNode.

        Args:
            service_name (str): Name of the AWS Service.

        Returns:
            ServiceNode: ServiceNode object representing an AWS Service
        """
        if service_name not in self._service_nodes_map:
            self._create_service_node(service_name)
        return self._service_nodes_map.get(service_name)

    def get_service_reader(self, service_name: str) -> ServiceReader:
        """Gets the ServiceReader obj from the ServiceNode obj.

        Args:
            service_name (str): Name of the AWS Service.

        Returns:
            ServiceReader: ServiceReader object with the read capabilities, tied to a ServiceNode.
        """
        service_node = self.get_service_node(service_name)
        if service_node:
            return service_node.get_service_reader()

    def read_operation(
        self,
        service_name: str,
        resource_node_name: str,
        operation_name: str,
        match_patterns: Optional[List[str]] = None,
        refresh: Optional[bool] = False,
    ) -> Union[dict, bool]:
        """Call the AWS API operation for the given `service_name`, `resource_node_name` and `operation_name` values.


        Args:
            service_name (str): AWS Service name.
            resource_node_name (str): AWS ResourceNode name
            operation_name (str): AWS Read opeartion name
            match_patterns (Optional[List[str]], optional): UNIX style patterns for generated required_parameters. Defaults to None.
            refresh (Optional[bool], optional): Force to re-read instead of returning the data from cache.. Defaults to False.

        Returns:
            Union[dict,bool]: Read Operation data, or False.
        """
        service_reader = self.get_service_reader(service_name)
        if service_reader:
            data = service_reader.read_operation(
                resource_node_name, operation_name, match_patterns, refresh
            )
            return data
        return False

    def read_resource_node(
        self,
        service_name: str,
        resource_node_name: str,
        match_patterns: Optional[List[str]] = None,
        refresh: Optional[bool] = False,
    ) -> Union[dict, bool]:
        """Reads all available Read operations of the given ResourceNode.

        Args:
            service_name (str): Name of the AWS Service
            resource_node_name (str): Name of the AWS Resource Node
            match_patterns (Optional[List[str]], optional): UNIX style patterns for generated required_parameters. Defaults to None.
            refresh (Optional[bool], optional): Force to re-read instead of returning the data from cache.. Defaults to False.

        Returns:
            Union[dict,bool]: Read ResourceNode data or False
        """
        service_reader = self.get_service_reader(service_name)
        if service_reader:
            data = service_reader.read_resource_node(
                resource_node_name, match_patterns, refresh
            )
            return data
        return False

    def get_available_service_names(self) -> List[str]:
        """Lists available AWS service names

        Returns:
            List[str]: Available service names for current `boto3.session.Session`
        """
        return get_all_available_services(self.boto3_session)



from botocore_utils import (
    get_shape_name,
    flatten_shape_to_its_non_collection_shape_and_target_paths,
    ShapeAndTargetPath,
)
from config import get_logger, BALCONY_RELATIONS_DIR
from utils import icompare_two_camel_case_words
from typing import List, Dict, Union
import json
import os
from dataclasses import dataclass


logger = get_logger(__name__)


@dataclass
class Relation:
    """Basic dataclass to hold the relation information."""
    service_name: str
    resource_node_name: str
    operation_name: str
    required_shape_name: str
    target_shape_name: str
    target_shape_type: str
    target_path: str

    def __str__(self):
        return f"{self.operation_name}:{self.required_shape_name}::{self.target_shape_name}"

    def __hash__(self):
        return hash(
            (
                self.service_name,
                self.resource_node_name,
                self.operation_name,
                self.required_shape_name,
                self.target_shape_name,
                self.target_shape_type,
                self.target_path,
            )
        )

    def __eq__(self, other):
        if other.__class__ is not self.__class__:
            return NotImplemented
        return (
            self.service_name,
            self.resource_node_name,
            self.operation_name,
            self.required_shape_name,
            self.target_shape_name,
            self.target_shape_type,
            self.target_path,
        ) == (
            other.service_name,
            other.resource_node_name,
            other.operation_name,
            other.required_shape_name,
            other.target_shape_name,
            other.target_shape_type,
            other.target_path,
        )


class RelationMap:
    """
    RelationMap represents a mapping `ParameterName` to `Relations List`.
    `generate_relation_map()` function generates the RelationMap dictionary for given ServiceNode.

    ServiceNode caches the RelationMap dictionary to a file, because calculating it
    requires a lot of loops.

    ```json title="Example RelationMap dictionary"
     {
        "PolicyArn": [
            {
                "service_name": "iam",
                "resource_node_name": "Policy",
                "required_shape_name": "PolicyArn",
                "target_shape_name": "Arn",
                "target_shape_type": "string",
                "operation_name": "ListPolicies",
                "target_path": "Policies[*].Arn"
            }
        ],
    }
    ```
    """

    def __init__(self, service_node):
        self.service_node = service_node
        self._relations_map: Dict = None


    def serialize_relations_map(self, relations_map: Dict) -> Dict:
        """Serializes the RelationMap dictionary to a json serializable dictionary.

        Args:
            relations_map (Dict): RelationMap dictionary

        Returns:
            Dict: Json serialized dictionary
        """
        serialized_relations_map = dict()
        for parameter_name, relations in relations_map.items():
            serialized_relations = list()
            for relation in relations:
                serialized_relations.append(relation.__dict__)
            serialized_relations_map[parameter_name] = serialized_relations
        return serialized_relations_map

    def deserialize_relation_map(self, serialized_relations_map: Dict) -> Dict:
        """Deserializes the RelationMap dictionary from a json serialized dictionary.

        Args:
            serialized_relations_map (Dict): Json serialized dictionary

        Returns:
            Dict: RelationMap dictionary
        """
        relations_map = dict()
        for parameter_name, relations in serialized_relations_map.items():
            deserialized_relations = list()
            for relation in relations:
                deserialized_relations.append(Relation(**relation))
            relations_map[parameter_name] = deserialized_relations
        return relations_map


    def get_relations_map(self, refresh: bool = False) -> Dict:
        """Tries the fetch the cached RelationMap from file,
        or generates and saves it to file.

        Args:
            refresh (bool, optional): Disables cache. Defaults to False.

        Returns:
            Dict: `RelationMap` dictionary. parameter_name->[relations,] mapping.
        """
        if self._relations_map is not None and not refresh:
            return self._relations_map

        loaded_relations_map = self.load_relations_from_file()
        if loaded_relations_map and not refresh:
            self._relations_map = self.deserialize_relation_map(loaded_relations_map)
            return self._relations_map

        generated_relations = self.generate_relation_map()
        self._relations_map = generated_relations
        self.save_relations_map_to_file(self.serialize_relations_map(self._relations_map))
        return self._relations_map

    def get_parameters_generated_relations(
        self, parameter_name: str, exclude_operation_name: str
    ) -> List[Dict]:
        """Checks the `RelationMap` dictionary for the key: `parameter_name`
        to get the generated relations for the given parameter name.

        Args:
            parameter_name (str): Parameter name to search for its relations
            exclude_operation_name (str): Exclude the given operation_name across found relation dicts

        Returns:
            List[Dict]: List of relations
        """
        relation_map = self.get_relations_map()
        relation_list = list()

        for map_parameter_name in relation_map.keys():
            # if compare_two_camel_case_words(map_parameter_name, parameter_name):
            if map_parameter_name.lower() == parameter_name.lower():
                relation_list.extend(relation_map.get(map_parameter_name, []))

        if relation_list:
            non_self_referencing_relations = [
                relation_obj
                for relation_obj in relation_list
                if relation_obj.operation_name != exclude_operation_name
            ]
            return non_self_referencing_relations

        return False

    def save_relations_map_to_file(self, relations_map: List[Dict] = None) -> None:
        """Saves the generated relations map to `{BALCONY_RELATIONS_DIR}/{service_name}.json`

        Args:
            relations_map (List[Dict], optional): _description_. Defaults to None.
        """

        directory = BALCONY_RELATIONS_DIR
        service_name = self.service_node.name
        filepath = os.path.join(directory, f"{service_name}.json")
        if relations_map is None:
            relations_map = self.get_relations_map()
        logger.debug(
            f"Caching relations of [bold green]{service_name}[/] to {filepath}. Use: [bold]balcony clear-cache[/] to remove."
        )
        with open(filepath, "w") as file:
            json.dump(relations_map, file, indent=2, default=str)

    def load_relations_from_file(self) -> Union[dict, bool]:
        directory = BALCONY_RELATIONS_DIR
        service_name = self.service_node.name
        filepath = os.path.join(directory, f"{service_name}.json")
        if not os.path.isfile(filepath):
            return False
        relations_map = None
        try:
            with open(filepath, "r") as file:
                relations_map = json.load(file)
            logger.debug(
                f"Read cached relations of [bold green]{service_name}[/] from {filepath}. Use: [bold]balcony clear-cache[/] to remove."
            )
        except:  # noqa
            return False
        return relations_map

    def _generate_resource_node_parameters_list(self, resource_nodes):
        _list = []
        for resource_node in resource_nodes:
            for operation_name in resource_node.operation_names:
                required_parameter_names = (
                    resource_node.get_required_parameter_names_from_operation_name(
                        operation_name
                    )
                )
                _list.append(
                    {
                        "resource_node": resource_node,
                        "operation_name": operation_name,
                        "required_parameter_names": required_parameter_names,
                    }
                )
        return _list

    def generate_relation_map(self, **kwargs) -> Dict[str, list]:  # noqa
        """Generates the required parameter name to its relations mapping.
        ```json title="RelationMap structure"
        {
            "ParameterName1": [{}, {} ],
            "ParameterName2": [{}, ]
        }
        ```

        Returns:
            Dict[str, list]: Generated parameter names to Relation dict list map
        """
        resource_nodes = self.service_node.get_resource_nodes()
        # Create a list of dicts for each resource_node, operation_name and required_parameter_names
        # to be able to calculate the all required parameter names befgore generating the relations
        operations_to_required_parameter_list = self._generate_resource_node_parameters_list(resource_nodes) # noqa

        # Get every unique required parameter names
        unique_required_shape_names = set()
        for _dict in operations_to_required_parameter_list:
            for r_param_name in _dict.get("required_parameter_names"):
                unique_required_shape_names.add(r_param_name)
        unique_required_shape_names_list = list(unique_required_shape_names)

        # create the relations map beforehand, and add empty sets to each required parameter name
        relations_map = {
            required_shape_name: set()
            for required_shape_name in unique_required_shape_names_list
        }
        for __dict in operations_to_required_parameter_list:
            # loop through all resource nodes and their operations
            resource_node = __dict.get("resource_node")
            operation_name = __dict.get("operation_name")
            # required_parameter_names=__dict.get('required_parameter_names')

            operation_model = resource_node.get_operation_model(operation_name)
            output_shape_member = operation_model.output_shape

            resource_nodes_extra_relations = resource_node.define_extra_relations()
            if resource_nodes_extra_relations:
                for extra_relation in resource_nodes_extra_relations:
                    relation_obj = extra_relation
                    # cast dictionaries as a Relation object
                    if type(extra_relation) == dict:
                        relation_obj = Relation(**extra_relation)
                    # make sure the relation object is of type Relation
                    assert type(relation_obj) == Relation
                    if relation_obj.required_shape_name not in relations_map:
                        # make sure there's a set for this required shape name in the map
                        relations_map[relation_obj.required_shape_name] = set()
                    relations_map[relation_obj.required_shape_name].add(relation_obj)

            # get a list of output shape's members, with their target path
            output_shape_and_target_paths = (
                flatten_shape_to_its_non_collection_shape_and_target_paths(output_shape_member)
            )
            # for each member shape (attr.) of this operations output shape
            # try to find it in the unique required parameter list
            for output_shape_and_target_path in output_shape_and_target_paths:
                # unpack the shape and target path
                output_shape_member = output_shape_and_target_path.shape
                target_path = output_shape_and_target_path.target_path
                output_shape_member_name = get_shape_name(output_shape_member)

                # look for a match between eac required shape name
                # and the current output's member's shapes name
                for required_shape_name in unique_required_shape_names_list:
                    # check both names with plurality and camelcase insensitivity
                    _match = icompare_two_camel_case_words(
                        output_shape_member_name, required_shape_name
                    )
                    # if there's a required shape name found in any member, add it to the relations map
                    if _match:
                        relation_obj = Relation(
                            service_name=self.service_node.name,
                            resource_node_name=resource_node.name,
                            operation_name=operation_name,
                            required_shape_name=required_shape_name,
                            target_shape_name=output_shape_member_name,
                            target_shape_type=output_shape_member.type_name,
                            target_path=target_path,
                        )
                        # we'll be adding to a set, so we don't need to worry about duplicates
                        relations_map[required_shape_name].add(relation_obj)

        return relations_map



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['balcony', 'balcony.custom_nodes']

package_data = \
{'': ['*'], 'balcony.custom_nodes': ['yamls/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'boto3>=1.24.80,<2.0.0',
 'inflect>=6.0.0,<7.0.0',
 'jmespath>=1.0.1,<2.0.0',
 'mkdocs-autorefs>=0.4.1,<0.5.0',
 'mkdocs-material>=8.5.7,<9.0.0',
 'mkdocstrings[python]>=0.19.0,<0.20.0',
 'pydantic>=1.10.7,<2.0.0',
 'rich>=12.5.1,<13.0.0',
 'typer>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['balcony = balcony.cli:run_app']}

setup_kwargs = {
    'name': 'balcony',
    'version': '0.0.77',
    'description': 'AWS API for humans',
    'long_description': "# balcony\n\nbalcony is a Python based CLI tool that simplifies the process of enumerating AWS resources.\n\nbalcony dynamically parses `boto3` library and analyzes required parameters for each operation. \n\nBy establishing relations between operations over required parameters, it's able to auto-fill them by reading the related operation beforehand.\n\nBy simply entering their name, balcony enables developers to easily list their AWS resources.\n\nIt uses _read-only_ operations, it does not take any action on the used AWS account.\n\n### [Go to Documentation Website](https://oguzhan-yilmaz.github.io/balcony/quickstart)\n\n## Features\n\n### List available AWS Services \nUse `balcony aws` to see every AWS service available.\n\n![](https://github.com/oguzhan-yilmaz/balcony/blob/main/docs/visuals/aws-services-list.gif)\n\n\n### List Resource Nodes of an AWS Service \nUse `balcony aws <service-name>` to see every Resource Node of a service.\n\n![](https://github.com/oguzhan-yilmaz/balcony/blob/main/docs/visuals/resource-node-list.gif)\n\n\n### Reading a Resource Node \nUse `balcony aws <service-name> <resource-node>` to read operations of a Resource Node.\n\n![](https://github.com/oguzhan-yilmaz/balcony/blob/main/docs/visuals/reading-a-resource-node.gif)\n\n\n### Documentation and Input & Output of Operations\nUse the `--list`, `-l` flag to print the given AWS API Operations documentation, input & output structure. \n \n\n![](https://github.com/oguzhan-yilmaz/balcony/blob/main/docs/visuals/list-option.gif)\n\n\n### Enable Debug messages \nUse the `--debug`, `-d` flag to see what's going on under the hood!\n\n![](https://github.com/oguzhan-yilmaz/balcony/blob/main/docs/visuals/debug-messages.gif)\n",
    'author': 'Oguzhan Yilmaz',
    'author_email': 'oguzhanylmz271@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
